# ##############################################################################
#  This file is part of Interdiode                                             #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <matthieu.gallet@19pouces.net>           #
#  All Rights Reserved                                                         #
#                                                                              #
# ##############################################################################
"""Root URLs provided by df_config.

By default, register URLs for the admin site, `jsi18n`, static and media files, favicon and robots.txt.
If DjangoDebugToolbar is present, then its URL is also registered.

"""

from django.conf import settings
from django.conf.urls import include
from django.urls import path, re_path
from django.utils.module_loading import autodiscover_modules, import_string
from django.views.i18n import JavaScriptCatalog
from django.views.static import serve

from df_config.utils import get_view_from_string


class RootUrls(list):
    """Root URLs provided by df_config.

    By default, register URLs for the admin site, `jsi18n`, static and media files, favicon and robots.txt.
    If DjangoDebugToolbar is present, then its URL is also registered.

    """

    def load_all_defaults(self):
        """Load all default URLs."""
        self.load_file_servers()
        self.load_wellknown_urls()
        self.load_settings_views()
        self.load_auth_urls()
        self.load_admin_site()
        self.load_prometheus()
        self.load_debug_toolbar()
        self.load_websockets()
        self.load_index_view()

    def load_file_servers(self):
        """Register URLs for static and media files."""
        if getattr(settings, "DF_JS_CATALOG_VIEWS", []):
            catalog_view = JavaScriptCatalog.as_view(
                packages=settings.DF_JS_CATALOG_VIEWS
            )
            self.append(
                path("jsi18n/", catalog_view, name="jsi18n"),
            )
        if settings.MEDIA_URL:
            self.append(
                path(
                    "%s<path:path>" % settings.MEDIA_URL[1:],
                    serve,
                    name="serve_media",
                    kwargs={"document_root": settings.MEDIA_ROOT},
                )
            )
        if settings.STATIC_URL:
            self.append(
                path(
                    "%s<path:path>" % settings.STATIC_URL[1:],
                    serve,
                    name="serve_static",
                    kwargs={"document_root": settings.STATIC_ROOT},
                )
            )

    def load_wellknown_urls(self):
        """Provide standard static URLs that should be exposed by every site."""
        self.extend(common_static_urls())

    def load_settings_views(self):
        """Register URLs for views defined in the settings."""
        if getattr(settings, "DF_URL_CONF", None):
            try:
                extra_urls = import_string(settings.DF_URL_CONF)
                self.extend(extra_urls)
            except ModuleNotFoundError:
                pass

    def load_auth_urls(self):
        """Register URLs for authentication."""
        if getattr(settings, "USE_ALL_AUTH", False):
            from allauth.account.views import login

            self.extend(
                [
                    path("admin/login/", login),
                    path("accounts/", include("allauth.urls")),
                ]
            )
        else:
            self.extend([path("auth/", include("django.contrib.auth.urls"))])

    def load_admin_site(self):
        """Register the admin site if it is enabled."""
        if getattr(settings, "DF_ADMIN_SITE", None):
            admin_site = import_string(settings.DF_ADMIN_SITE)
            autodiscover_modules("admin", register_to=admin_site)
            self.append(path("admin/", include(admin_site.urls[:2])))

    def load_prometheus(self):
        """Register the django-prometheus URL if it is enabled."""
        if (
            getattr(settings, "USE_PROMETHEUS", False)
            and getattr(settings, "PROMETHEUS_URL_PREFIX", None) is not None
        ):
            self.append(
                path(settings.PROMETHEUS_URL_PREFIX, include("django_prometheus.urls"))
            )

    def load_debug_toolbar(self):
        """Register the debug toolbar if it is enabled."""
        if getattr(settings, "DEBUG") and getattr(settings, "USE_DEBUG_TOOLBAR"):
            import debug_toolbar

            self.append(path("__debug__/", include(debug_toolbar.urls)))

    @staticmethod
    def load_websockets():
        """Load the websockets thins when df_websockets is used."""
        if getattr(settings, "USE_WEBSOCKETS", False):
            try:
                from df_websockets.load import load_celery

                load_celery()
            except ImportError:
                # noinspection PyUnresolvedReferences
                from df_websockets.tasks import import_signals_and_functions

                import_signals_and_functions()

    def load_index_view(self):
        """Register a view for the index page."""
        if getattr(settings, "DF_INDEX_VIEW", None):
            self.append(
                path("", get_view_from_string(settings.DF_INDEX_VIEW), name="index")
            )

    def prefixed(self):
        """Add a prefix to all URLs."""
        url_prefix = (getattr(settings, "URL_PREFIX", "") or "")[1:]
        if url_prefix:
            return [re_path("^" + url_prefix, include(self))]
        return self


def common_static_urls():
    """Provide standard static URLs that should be exposed by every site."""
    values = (
        "robots.txt",
        "apple-touch-icon.png",
        "apple-touch-icon-precomposed.png",
        "favicon.ico",
    )
    return [
        path(
            filename,
            serve,
            kwargs={
                "document_root": settings.STATIC_ROOT,
                "path": "favicon/%s" % filename,
            },
        )
        for filename in values
    ]


_urlpatterns = RootUrls()
_urlpatterns.load_all_defaults()
urlpatterns = _urlpatterns.prefixed()



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
"""Some utility functions."""
import argparse
import datetime
import mimetypes
import os
import re
from email.utils import mktime_tz, parsedate_tz
from importlib import metadata
from importlib.util import find_spec
from typing import Iterable, Optional, Set, Tuple
from urllib.parse import quote

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseNotModified,
    StreamingHttpResponse,
)
from django.utils.http import http_date
from django.utils.module_loading import import_string


class RemovedInDjangoFloor200Warning(DeprecationWarning):
    """Used for displaying functions or modules that will be removed in a near future."""

    pass


def ensure_dir(path, parent=True):
    """Ensure that the given directory exists.

    :param path: the path to check
    :param parent: only ensure the existence of the parent directory

    """
    dirname = os.path.dirname(path) if parent else path
    os.makedirs(dirname, exist_ok=True)
    return path


def is_package_present(package_name):
    """Return True is the `package_name` package is present in your current Python environment."""
    return find_spec(package_name) is not None


def remove_arguments_from_help(parser: argparse.ArgumentParser, arguments: Set[str]):
    """Remove the arguments from help message."""
    # noinspection PyProtectedMember
    for action in parser._actions:
        if arguments & set(action.option_strings):
            action.help = argparse.SUPPRESS


def guess_version(defined_settings):
    """Guess the project version.

    Expect an installed version (findable with pkg_resources) or __version__ in `your_project/__init__.py`.
    If not found, return "1.0.0".

    :param defined_settings: all already defined settings (dict)
    :type defined_settings: :class:`dict`
    :return: should be something like `"1.2.3"`
    :rtype: :class:`str`
    """
    try:
        return metadata.version(defined_settings["DF_MODULE_NAME"])
    except metadata.PackageNotFoundError:
        pass
    try:
        return import_string(f"{defined_settings['DF_MODULE_NAME']}.__version__")
    except ImportError:
        return "1.0.0"


def get_view_from_string(view_as_str):
    """Return a view from the given string."""
    try:
        view = import_string(view_as_str)
    except ImportError:
        raise ImproperlyConfigured(f"Unable to import {view_as_str}")
    if hasattr(view, "as_view") and isinstance(view, type):
        return view.as_view()
    elif callable(view) and not isinstance(view, type):
        return view
    raise ImproperlyConfigured(
        f'{view} is not callable and does not have an "as_view" attribute'
    )


class ChunkReader:
    """Read a file object in chunks of the given size.

    Return an iterator of data

    :param fileobj:
    :param chunk_size: max size of each chunk
    :type chunk_size: `int`
    """

    def __init__(self, fileobj, chunk_size=32768):
        """Initialize the ChunkReader object."""
        self.fileobj = fileobj
        self.chunk_size = chunk_size

    def __iter__(self):
        """Iterate over the underlying file object."""
        for data in iter(lambda: self.fileobj.read(self.chunk_size), b""):
            yield data

    def close(self):
        """Close the underlying file object."""
        self.fileobj.close()


class RangedChunkReader(ChunkReader):
    """Read the given chunks of the file."""

    def __init__(self, fd, ranges: Iterable[Tuple[int, int]], chunk_size=32768):
        """Initialize a RangedChunkReader."""
        super().__init__(fd, chunk_size=chunk_size)
        self.ranges = ranges

    def __iter__(self):
        """Read the required chunks of the file."""
        for start, end in self.ranges:
            self.fileobj.seek(start)
            while start <= end:
                size = min(self.chunk_size, end - start + 1)
                if size <= 0:
                    break
                yield self.fileobj.read(size)
                start += size


mimetypes.init()


_range_re = re.compile(r"^bytes=(\d*-\d*(?:,\s*\d*-\d*)*)$")
_if_modified_since_re = re.compile(
    r"^([^;]+)(;\s*length=([1-9]\d*))?$", flags=re.IGNORECASE
)


def was_modified_since(
    header: Optional[str] = None, mtime: float = 0.0, size: int = None
) -> bool:
    """Return true is something has been modified since the user last downloaded it.

    Doc: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-Modified-Since

    :param header: Value of the If-Modified-Since header.  If None, return True.
    :param mtime: Modification time of the item we're talking about.
    :param size: Size of the item we're talking about (can be provided in the header).
    """
    if header is None:
        return True
    matcher = _if_modified_since_re.match(header)
    if not matcher:
        return True
    try:
        header_date = parsedate_tz(matcher.group(1))
        if header_date is None:
            return True
        # noinspection PyTypeChecker
        header_mtime = mktime_tz(header_date)
        if mtime > header_mtime:
            return True
        header_len = matcher.group(3)
        if size is not None and header_len and int(header_len) != size:
            return True
    except (AttributeError, ValueError, OverflowError):
        return True
    return False


def send_file(
    request: HttpRequest,
    filepath: str,
    mimetype=None,
    force_download=False,
    attachment_filename: Optional[str] = None,
    chunk_size: int = 32768,
    etag: Optional[str] = None,
    expires: Optional[datetime.datetime] = None,
):
    """Send a local file. This is not a Django view, but a function that is called at the end of a view.

    If `settings.USE_X_SEND_FILE` (mod_xsendfile is a mod of Apache), then return an empty HttpResponse with the
    correct header. The file is directly handled by Apache instead of Python.
    If `settings.X_ACCEL_REDIRECT_ARCHIVE` is defined (as a list of tuple (directory, alias_url)) and filepath is
    in one of the directories, return an empty HttpResponse with the correct header.
    This is only available with Nginx.

    Otherwise, return a StreamingHttpResponse to avoid loading the whole file in memory.

    :param request: the original request (used to detect the "range" header)
    :param filepath: absolute path of the file to send to the client.
    :param mimetype: MIME type of the file (returned in the response header)
    :param force_download: always force the client to download the file.
    :param attachment_filename: filename used in the "Content-Disposition" header (when used)
    :param chunk_size: size of chunks for large files. Useful at least for unittests
    :param etag: ETag header to add to the response
    :param expires: expiration date of the file
    :rtype: :class:`django.http.response.StreamingHttpResponse` or :class:`django.http.response.HttpResponse`
    """
    if mimetype is None:
        (mimetype, encoding) = mimetypes.guess_type(filepath)
        if mimetype is None:
            mimetype = "text/plain"

    filepath = os.path.abspath(filepath)
    if_modified_since = request.META.get("HTTP_IF_MODIFIED_SINCE")
    if not os.path.isfile(filepath):
        return HttpResponse(
            status=404,
            content="File not found.",
            content_type="text/plain; charset=utf-8",
        )
    stats = os.stat(filepath)

    def send_response(r, cs: Optional[int] = None):
        r["Last-Modified"] = http_date(stats.st_mtime)
        if etag:
            r["ETag"] = str(etag)
        if expires:
            r["Expires"] = http_date(expires.timestamp())
        if cs is not None:
            r["Content-Length"] = str(cs)
        return r

    attachment_filename = attachment_filename or os.path.basename(filepath)
    range_matcher = _range_re.match(request.META.get("HTTP_RANGE", ""))
    ranges = []
    filesize = stats.st_size
    if not was_modified_since(if_modified_since, mtime=stats.st_mtime, size=filesize):
        return send_response(HttpResponseNotModified())
    if range_matcher:
        content_size = 0
        ranges_str = [x.strip() for x in range_matcher.group(1).split(",")]
        for range_str in ranges_str:
            start_str, sep, end_str = range_str.partition("-")
            end = int(end_str) if end_str else filesize - 1
            start = int(start_str) if start_str else filesize - end
            content_size += end - start + 1
            if end + 1 > filesize:
                response = HttpResponse(content_type=mimetype, status=416)
                response["Content-Range"] = f"bytes */{filesize}"
                return send_response(response)
            ranges.append((start, end))
    else:
        content_size = filesize
    if request.method == "HEAD":
        response = HttpResponse(content=b"", content_type=mimetype, status=200)
        return send_response(response, cs=content_size)

    response = None
    if settings.USE_X_SEND_FILE and not ranges:
        response = HttpResponse(content_type=mimetype)
        response["X-SENDFILE"] = filepath
    elif settings.X_ACCEL_REDIRECT and not ranges:
        for dirpath, alias_url in settings.X_ACCEL_REDIRECT:
            dirpath = os.path.abspath(dirpath)
            if filepath.startswith(dirpath):
                response = HttpResponse(content_type=mimetype)
                response["X-Accel-Redirect"] = os.path.join(
                    alias_url, os.path.relpath(filepath, dirpath)
                )
                break
    if response is None:
        try:
            fileobj = open(filepath, "rb")
        except OSError:
            return HttpResponse(status=40, content="Unable to open the file.")
        status = 200
        if ranges:
            file_content = RangedChunkReader(fileobj, ranges, chunk_size=chunk_size)
            if len(ranges) == 1:
                status = 206
        else:
            file_content = ChunkReader(fileobj, chunk_size=chunk_size)
        response = StreamingHttpResponse(
            file_content, content_type=mimetype, status=status
        )
        if len(ranges) == 1:
            response["Content-Range"] = "bytes %d-%d/%d" % (
                ranges[0][0],
                ranges[0][1],
                filesize,
            )
    encoded_filename = quote(attachment_filename, encoding="utf-8")
    header = "attachment" if force_download else "inline"
    if encoded_filename == attachment_filename:
        response["Content-Disposition"] = '{1}; filename="{0}"'.format(
            encoded_filename, header
        )
    else:
        response["Content-Disposition"] = "{1}; filename*=\"UTF-8''{0}\"".format(
            encoded_filename, header
        )
    return send_response(response, cs=content_size)



"""Allow to create different dynamic settings from a single URL."""

import re
import urllib.parse
from importlib.metadata import PackageNotFoundError, version
from typing import Callable, Dict, List, Optional, Tuple, Union

from django.core.exceptions import ImproperlyConfigured

from df_config.config.dynamic_settings import DynamicSettting


class Attribute(DynamicSettting):
    """Dynamic setting which is an attribute of a URL string."""

    def __init__(
        self,
        url_setting,
        attribute: Callable[[], Optional[Union[str, bool]]],
        default=None,
    ):
        """Initialize the object."""
        super().__init__(attribute)
        self.url_setting: URLSetting = url_setting
        self.default = default

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Get the final value of the attribute."""
        self.url_setting.load(merger)
        if self.url_setting.parsed_urls:
            value = self.value()
        else:
            value = self.default
        return merger.analyze_raw_value(value, provider_name, setting_name)

    def __repr__(self):
        """Represent this object as a string."""
        method = self.value.__name__[:-1]
        default = self.default
        setting = (
            f"{self.url_setting.__class__.__name__}({self.url_setting.setting_name!r})"
        )
        return f"{setting}.{method}(default={default!r})"


class URLSetting:
    """Represents an URL setting, such that each URL component can be used as a DynamicSetting."""

    ENGINES = {}
    REQUIREMENTS = {}
    SCHEME_ALIASES = {}
    SCHEMES: Dict[str, Tuple[int, bool, bool]] = {
        "amqp": (5672, False, False),
        "file": (None, False, False),
        "http": (80, False, False),
        "https": (443, True, False),
        "memcache": (11211, False, False),
        "mariadb": (3306, False, False),
        "mysql": (3306, False, False),
        "oracle": (1521, False, False),
        "postgres": (5432, False, False),
        "redis": (6379, False, False),
        "rediss": (6379, True, False),
        "smtp": (25, False, False),
        "smtp+tls": (487, False, True),
        "smtps": (465, True, False),
        "sqlite": (None, False, False),
        "sqlite3": (None, False, False),
    }  # (port, SSL ?, StartTLS ?)

    def __init__(
        self,
        setting_name: str = None,
        required: List[str] = None,
        url: Optional[str] = None,
        split_char: str = ",",
        split_netloc_only: bool = True,
    ):
        """Initialize the object.

        :param setting_name: The name of the Django setting that contains the URL string.
        :param required: A list of setting names that are required to be loaded before this one.
        :param url: The URL string to be parsed.
        :param split_char: A character to split the raw URL string into multiple URLs (useful for DB clusters).
        :param split_netloc_only: If True, scheme, auth, params and path parts must be identical across URLs.


        """
        self.split_char = split_char
        self.split_netloc_only = split_netloc_only
        if not url:
            self._url_str = None
            self.parsed_urls = None
            self._loaded = False
        else:
            self._url_str = url
            if self.split_char != "":
                self.parsed_urls = [
                    urllib.parse.urlparse(x)
                    for x in self._url_str.split(self.split_char)
                ]
            else:
                self.parsed_urls = [urllib.parse.urlparse(self._url_str)]
            self._loaded = True
        self.setting_name = setting_name
        self.required = required or []

    def __repr__(self):
        """Represent this object as a string."""
        return f"{self.__class__.__name__}('{self.setting_name}')"

    def __str__(self):
        """Return the loaded URL string."""
        return self._url_str if self._url_str is not None else ""

    def load(self, merger):
        """Get the URL string for the merger, parse it and load it to extract its components."""
        if self._loaded or not self.setting_name:
            return
        for required in self.required:
            merger.get_setting_value(required)
        value = merger.get_setting_value(self.setting_name)
        if value is self or not value:
            self._url_str = None
            self.parsed_urls = None
        elif self.split_char != "":
            self._url_str = value
            self.parsed_urls = [
                urllib.parse.urlparse(x) for x in self._url_str.split(self.split_char)
            ]
        else:
            self._url_str = value
            self.parsed_urls = [urllib.parse.urlparse(self._url_str)]
        self._loaded = True

    def hostname(self, default="localhost"):
        """Return a DynamicSetting that represents the hostname."""
        return Attribute(self, self.hostname_, default=default)

    def hostname_(self) -> Optional[str]:
        """Return the hostname."""
        if self.parsed_urls is None:
            return None
        return self.split_char.join(x.hostname for x in self.parsed_urls)

    def netloc(self, default="localhost"):
        """Return a DynamicSetting that represents the netloc."""
        return Attribute(self, self.netloc_, default=default)

    def netloc_(self) -> Optional[str]:
        """Return the netloc."""
        if self.parsed_urls is None:
            return None
        return self.split_char.join(x.netloc for x in self.parsed_urls)

    def params(self, default=""):
        """Return a DynamicSetting that represents the params."""
        return Attribute(self, self.params_, default=default)

    def params_(self) -> Optional[str]:
        """Return the params part of the URL."""
        return self.get_attribute_("params")

    def get_attribute_(self, attr_name) -> Optional[str]:
        """Return the params part of the URL."""
        if self.parsed_urls is None:
            return None
        elif self.split_netloc_only:
            if len({getattr(x, attr_name) for x in self.parsed_urls}) > 1:
                raise ImproperlyConfigured(
                    f"{attr_name} must be identical across URLs '{self._url_str}'."
                )
            return getattr(self.parsed_urls[0], attr_name)
        return self.split_char.join(getattr(x, attr_name) for x in self.parsed_urls)

    def password(self, default=None):
        """Return a DynamicSetting that represents the password."""
        return Attribute(self, self.password_, default=default)

    def password_(self) -> Optional[str]:
        """Return the user password."""
        return self.get_attribute_("password")

    def path(self, default=""):
        """Return a DynamicSetting that represents the path of the URL."""
        return Attribute(self, self.path_, default=default)

    def path_(self) -> Optional[str]:
        """Return the path of the URL."""
        return self.get_attribute_("path")

    def port(self, default=None):
        """Return a DynamicSetting that represents the port of the URL."""
        return Attribute(self, self.port_, default=default)

    def port_(self) -> Optional[str]:
        """Return the port of the URL."""
        if self.parsed_urls is None:
            return None
        ports = []
        for parsed_url in self.parsed_urls:
            if parsed_url.port:
                ports.append(str(parsed_url.port))
            else:
                s = parsed_url.scheme.lower()
                s = self.SCHEME_ALIASES.get(s, s)
                ports.append(str(self.SCHEMES[s][0]))
        return self.split_char.join(ports)

    def query(self, default=""):
        """Return a DynamicSetting that represents the query string from the URL."""
        return Attribute(self, self.query_, default=default)

    def query_(self) -> Optional[str]:
        """Return the query string from the URL."""
        return self.get_attribute_("query")

    def scheme(self, default=None):
        """Return a DynamicSetting that represents the URL scheme."""
        return Attribute(self, self.scheme_, default=default)

    def scheme_(self) -> Optional[str]:
        """Return the URL scheme."""
        if self.parsed_urls is None:
            return None
        schemes = [x.scheme.lower() for x in self.parsed_urls]
        schemes = [self.SCHEME_ALIASES.get(x, x) for x in schemes]
        if self.split_netloc_only:
            if len(set(schemes)) > 1:
                raise ImproperlyConfigured(
                    f"Scheme must be identical across URLs '{self._url_str}'."
                )
            return schemes[0]
        return self.split_char.join(schemes)

    def username(self, default=None):
        """Return a DynamicSetting that represents the username."""
        return Attribute(self, self.username_, default=default)

    def username_(self):
        """Return the username, if defined in the URL string."""
        return self.get_attribute_("username")

    def database(self, default=None):
        """Return a DynamicSetting that represents the database name."""
        return Attribute(self, self.database_, default=default)

    def database_(self):
        """Return the database name."""
        v = self.get_attribute_("path")
        if not v or not (matcher := re.match(r"/([^/]+)", v)):
            return None
        return matcher.group(1)

    def port_int(self, default=None):
        """Return an integer value for the port (the default one if not specified)."""
        return Attribute(self, self.port_int_, default=default)

    def port_int_(self) -> Optional[int]:
        """Return the port number for the port (the default one if not specified)."""
        if not self.parsed_urls:
            return None
        elif len(self.parsed_urls) > 1:
            raise ImproperlyConfigured("Multiple URLs, cannot return a single port.")
        parsed_url = self.parsed_urls[0]
        if parsed_url.port:
            return parsed_url.port
        s = parsed_url.scheme.lower()
        s = self.SCHEME_ALIASES.get(s, s)
        return self.SCHEMES[s][0]

    def use_tls(self, default=False):
        """Return a dynamic setting for the TLS usage."""
        return Attribute(self, self.use_tls_, default=default)

    def use_tls_(self):
        """Return True is TLS is used."""
        if not self.parsed_urls:
            return False
        elif len(self.parsed_urls) > 1:
            raise ImproperlyConfigured("Multiple URLs, cannot return a single value.")
        s = self.scheme_()
        return self.SCHEMES[s][2]

    def use_ssl(self, default=False):
        """Return a dynamic setting for the SSL usage."""
        return Attribute(self, self.use_ssl_, default=default)

    def use_ssl_(self):
        """Return True if SSL is used."""
        if not self.parsed_urls:
            return False
        elif len(self.parsed_urls) > 1:
            raise ImproperlyConfigured("Multiple URLs, cannot return a single value.")
        s = self.scheme_()
        return self.SCHEMES[s][1]

    def engine(self, default=None):
        """Return a dynamic setting for the database engine."""
        return Attribute(self, self.engine_, default=default)

    def engine_(self):
        """Return the engine name from the URL scheme."""
        if not self.parsed_urls:
            return None
        scheme = self.parsed_urls[0].scheme.lower()
        return self.normalize_engine(scheme)

    @classmethod
    def normalize_engine(cls, scheme: str):
        """Return a normalized Django engine name from a URL scheme string."""
        scheme = cls.SCHEME_ALIASES.get(scheme, scheme)
        engine = cls.ENGINES.get(scheme, scheme)
        requirements = cls.REQUIREMENTS.get(engine, [])
        found = False
        for req in requirements:
            try:
                version(req)
                found = True
            except PackageNotFoundError:
                pass
        if not found and requirements:
            from df_config.checks import missing_package, settings_check_results

            settings_check_results.append(
                missing_package("/".join(requirements), f" to use {engine}.")
            )
        return engine


class DatabaseURL(URLSetting):
    """Guess the database engines from a URL string."""

    ENGINES = {
        "mysql": "django.db.backends.mysql",
        "mariadb": "django.db.backends.mysql",
        "oracle": "django.db.backends.oracle",
        "postgres": "django.db.backends.postgresql",
        "sqlite3": "django.db.backends.sqlite3",
    }
    SCHEME_ALIASES = {
        "psql": "postgres",
        "postgresql": "postgres",
        "sqlite": "sqlite3",
    }
    REQUIREMENTS = {
        "django.db.backends.postgresql": ["psycopg2-binary", "psycopg2"],
        "django.db.backends.oracle": ["cx_Oracle"],
        "django.db.backends.mysql": ["mysqlclient", "pymysql"],
    }


class RedisURL(URLSetting):
    """Specialized class for Redis URLs, since databases are identified by a integer."""

    def database_(self) -> Optional[int]:
        """Extract a valid database number for Redis connections."""
        v = self.get_attribute_("path")
        if not v or not (matcher := re.match(r"/(0|[1-9]\d*)$", v)):
            return None
        return int(matcher.group(1))



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
"""Settings for database and cache backends."""
from typing import List, Set
from urllib.parse import ParseResult, urlencode, urlparse

from django.core.exceptions import ImproperlyConfigured
from django.utils import version

from df_config import utils
from df_config.config.dynamic_settings import Directory
from df_config.config.url import DatabaseURL

prometheus_engines = {
    "django.db.backends.sqlite3": "django_prometheus.db.backends.sqlite3",
    "django.db.backends.postgresql": "django_prometheus.db.backends.postgresql",
    "django.db.backends.mysql": "django_prometheus.db.backends.mysql",
    "django.core.cache.backends.filebased.FileBasedCache": "django_prometheus.cache.backends.filebased.FileBasedCache",
    "django.core.cache.backends.locmem.LocMemCache": "django_prometheus.cache.backends.locmem.LocMemCache",
    "django.core.cache.backends.memcached.PyLibMCCache": "django_prometheus.cache.backends.memcached.PyLibMCCache",
    "django.core.cache.backends.memcached.PyMemcacheCache": "django_prometheus.cache.backends.memcached.PyMemcacheCache",
    "django.core.cache.backends.redis.RedisCache": "django_prometheus.cache.backends.redis.RedisCache",
}


def databases(settings_dict):
    """Build a complete DATABASES setting for the default database.

    When django-prometheus is used, the engine is replaced by the matching prometheus engine.
    """
    engine = DatabaseURL.normalize_engine(settings_dict["DATABASE_ENGINE"])
    if settings_dict["USE_PROMETHEUS"]:
        engine = prometheus_engines.get(engine, engine)
    hosts, ports = None, None
    host = settings_dict["DATABASE_HOST"]
    if isinstance(host, str):
        hosts = host.split(",")
    port = settings_dict["DATABASE_PORT"]
    if isinstance(port, int):
        port = str(port)
    if isinstance(port, str):
        ports = port.split(",")
    if hosts and ports and len(ports) != len(hosts):
        raise ImproperlyConfigured(
            "DATABASE_HOST and DATABASE_PORT must have the same number of elements."
        )
    default = {
        "ENGINE": engine,
        "NAME": settings_dict["DATABASE_NAME"],
        "USER": settings_dict["DATABASE_USER"],
        "OPTIONS": settings_dict["DATABASE_OPTIONS"],
        "PASSWORD": settings_dict["DATABASE_PASSWORD"],
        "HOST": host,
        "PORT": port,
        "CONN_MAX_AGE": settings_dict["DATABASE_CONN_MAX_AGE"],
        "CONN_HEALTH_CHECKS": bool(settings_dict["DATABASE_CONN_MAX_AGE"]),
    }
    return {"default": default}


databases.required_settings = [
    "DATABASE_ENGINE",
    "DATABASE_NAME",
    "DATABASE_USER",
    "DATABASE_OPTIONS",
    "DATABASE_PASSWORD",
    "DATABASE_HOST",
    "DATABASE_PORT",
    "DATABASE_CONN_MAX_AGE",
    "USE_PROMETHEUS",
]


class RedisSmartSetting:
    """Handle values required for Redis configuration, as well as Heroku's standard environment variables.

    Can be used as :class:`df_config.config.dynamic_settings.CallableSetting`.
    """

    _config_values = ["PROTOCOL", "HOST", "PORT", "DB", "PASSWORD", "USERNAME"]

    def __init__(
        self,
        prefix="",
        env_variable="REDIS_URL",
        fmt="url",
        extra_values=None,
        only_redis: bool = True,
    ):
        """Build Redis connection parameters from a set of settings.

        These settings are:
        - "{prefix}PROTOCOL",
        - "{prefix}HOST",
        - "{prefix}PORT",
        - "{prefix}DB",
        - "{prefix}USERNAME",
        - "{prefix}PASSWORD".

        :param prefix: prefix of all settings
        :param env_variable: if this environment variable is present, override given settings
        :param fmt: output format:
            "url": redis://:password@host:port/db
            "dict": {"host": "host", "password": "password" or None, "port": port, "db": db}
            "channels": expected by django-channels
        :param extra_values: added to the output format
        """
        self.fmt = fmt
        self.prefix = prefix
        self.env_variable = env_variable
        self.only_redis = only_redis
        self.config_values = list(self._config_values)
        self.required_settings = [prefix + x for x in self.config_values]
        self.extra_values = extra_values

    def __call__(self, settings_dict):
        """Return the redis setting."""
        values = {x: settings_dict[self.prefix + x] for x in self.config_values}
        values["AUTH"] = ""
        if values["PASSWORD"]:
            values["AUTH"] = "%s:%s@" % (values["USERNAME"] or "", values["PASSWORD"])
        if self.fmt == "url":
            url = "%(PROTOCOL)s://%(AUTH)s%(HOST)s:%(PORT)s/%(DB)s" % values
            if self.extra_values:
                url += "?" + urlencode(self.extra_values)
            return url
        elif self.fmt == "channels":
            url = "%(PROTOCOL)s://%(AUTH)s%(HOST)s:%(PORT)s/%(DB)s" % values
            config = {
                "address": url,
                "password": values["PASSWORD"] or None,
                "db": int(values["DB"] or 0),
            }
            if self.extra_values:
                config.update(self.extra_values)
            # noinspection PyUnresolvedReferences
            result = {
                "BACKEND": "channels_redis.core.RedisChannelLayer",
                "CONFIG": {
                    "hosts": [config],
                    "capacity": 5000,
                    "expiry": 10,
                },
            }
            return result
        elif self.fmt == "dict":
            result = {
                "host": values["HOST"] or "localhost",
                "port": int(values["PORT"] or 6379),
                "db": int(values["DB"] or 0),
                "password": values["PASSWORD"] or None,
            }
            if self.extra_values:
                result.update(self.extra_values)
            return result
        raise ValueError("Unknown RedisSmartSetting format '%s'" % self.fmt)

    def __repr__(self):
        """Return a representation of the redis setting."""
        p = self.prefix
        if self.prefix.endswith("REDIS_"):
            p = self.prefix[:-6]
        return "%s.%sredis_%s" % (self.__module__, p.lower(), self.fmt)


cache_redis_url = RedisSmartSetting(prefix="CACHE_", fmt="url")
celery_broker_url = RedisSmartSetting(prefix="CELERY_", fmt="url", only_redis=False)
celery_result_url = RedisSmartSetting(
    prefix="CELERY_RESULT_", fmt="url", only_redis=False
)
session_redis_dict = RedisSmartSetting(
    prefix="SESSION_REDIS_", fmt="dict", extra_values={"prefix": "session"}
)
websocket_redis_dict = RedisSmartSetting(prefix="WEBSOCKET_REDIS_", fmt="dict")
websocket_redis_channels = RedisSmartSetting(prefix="WEBSOCKET_REDIS_", fmt="channels")


def cache_setting(settings_dict):
    """Automatically compute cache settings.

    Four caches are defined:

      * `base` that uses the CACHE_URL setting with redis, rediss or memcache protocols as soon as possible,
      * `locmem`: always uses the local memory, hence destroyed when the process is killed,
      * `cached`: "locmem" when DEBUG is true, `base` else,
      * `default`: "dummy" when DEBUG is true,`base` else.

    The rational is that developping (DEBUG=True) requires sometimes actually no cache at all
    (we do not want HTML caching) and sometimes a cache (the authentication cannot work with no cache at all).

    :param settings_dict:
    :return:
    """
    cache_url: str = settings_dict["CACHE_URL"]
    parsed_urls: List[ParseResult] = [urlparse(x) for x in cache_url.split(",")]
    schemes: Set[str] = {x.scheme.lower() for x in parsed_urls}
    django_version = version.get_complete_version()
    backend = "django.core.cache.backends.locmem.LocMemCache"
    prometheus_engines_ = {}
    if settings_dict.get("USE_PROMETHEUS", False):
        prometheus_engines_ = prometheus_engines
    locmem = {
        "BACKEND": prometheus_engines_.get(backend, backend),
        "LOCATION": "unique-snowflake",
    }

    backend = "django.core.cache.backends.dummy.DummyCache"
    dummy = {"BACKEND": prometheus_engines_.get(backend, backend)}
    actual = locmem
    if django_version >= (4, 0) and schemes.issubset({"redis", "rediss"}):
        backend = "django.core.cache.backends.redis.RedisCache"
        actual = {
            "BACKEND": prometheus_engines_.get(backend, backend),
            "LOCATION": [x.geturl() for x in parsed_urls],
        }
    elif schemes.issubset({"redis", "rediss"}):
        if utils.is_package_present("django_redis"):
            # noinspection PyUnresolvedReferences
            actual = {
                "BACKEND": "django_redis.cache.RedisCache",
                "LOCATION": [x.geturl() for x in parsed_urls],
                "OPTIONS": {"CLIENT_CLASS": "django_redis.client.DefaultClient"},
            }
        else:
            raise ImproperlyConfigured(
                "Please install 'django-redis' package to use Redis cache."
            )
    elif schemes == {"memcache"}:
        if django_version >= (3, 2) and utils.is_package_present("pymemcache"):
            backend = "django.core.cache.backends.memcached.PyMemcacheCache"
        elif django_version >= (3, 2) and utils.is_package_present("pylibmc"):
            backend = "django.core.cache.backends.memcached.PyLibMCCache"
        else:
            raise ImproperlyConfigured(
                "Please install 'pylibmc' package before using memcache engine."
            )
        info = [(x.hostname or "localhost", x.port or 11211) for x in parsed_urls]
        actual = {
            "BACKEND": prometheus_engines_.get(backend, backend),
            "LOCATION": [f"{x[0]}:{x[1]}" for x in info],
        }
    elif schemes == {"file"}:
        backend = "django.core.cache.backends.filebased.FileBasedCache"
        if len(parsed_urls) > 1:
            raise ImproperlyConfigured(
                "CACHE_URL with 'file' scheme must contain only one URL."
            )
        actual = {
            "BACKEND": prometheus_engines_.get(backend, backend),
            "LOCATION": Directory(parsed_urls[0].path),
        }
    elif cache_url:
        raise ImproperlyConfigured(
            "CACHE_URL must be a list of URLs with the 'file', 'redis', 'rediss' or 'memcache' scheme."
        )
    default = dummy if settings_dict["DEBUG"] else actual
    cached = locmem if settings_dict["DEBUG"] else actual
    return {"default": default, "locmem": locmem, "base": actual, "cached": cached}


cache_setting.required_settings = [
    "DEBUG",
    "CACHE_URL",
    "USE_PROMETHEUS",
]



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
"""Define a new authentication backend that add the newly created user to the selected groups."""
import logging

from django.conf import settings
from django.contrib.auth.backends import RemoteUserBackend
from django.contrib.auth.models import Group
from django.http import HttpRequest
from django.utils.functional import cached_property

logger = logging.getLogger("django.request")

_CACHED_GROUPS = {}


class DefaultGroupsRemoteUserBackend(RemoteUserBackend):
    """Add groups to new users.

    Based on :class:`django.contrib.auth.backends.RemoteUserBackend`.
    Only overrides the `configure_user` method to add the required groups.
    """

    @property
    def create_unknown_user(self):
        """Return True if the user should be automatically created."""
        return getattr(settings, "DF_ALLOW_USER_CREATION", True)

    @cached_property
    def ldap_backend(self):
        """Return the LDAP backend if it is enabled."""
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        from django_auth_ldap.backend import LDAPBackend

        return LDAPBackend()

    def authenticate(self, *args, **kwargs):
        """Authenticate the user and update it if needed."""
        remote_user = kwargs.get("remote_user")
        if (
            remote_user
            and getattr(settings, "AUTH_LDAP_SERVER_URI", None)
            and getattr(settings, "AUTH_LDAP_ALWAYS_UPDATE_USER", None)
        ):
            user = self.ldap_backend.populate_user(remote_user)
            if user:
                return user
        return super().authenticate(*args, remote_user)

    def configure_user(self, request, user=None, created=True):
        """Configure a user after creation and returns the updated user.

        By default, returns the user unmodified; only add it to the default group.
        """
        if user is None and not isinstance(request, HttpRequest):  # for compatibility
            user = request
        if user is None:
            return
        for group_name in getattr(settings, "DF_DEFAULT_GROUPS", []):
            if group_name not in _CACHED_GROUPS:
                _CACHED_GROUPS[group_name] = Group.objects.get_or_create(
                    name=str(group_name)
                )[0]
            user.groups.add(_CACHED_GROUPS[group_name])
        return user
